/*
 *  debugserver.h: handle usecode debugging messages from client
 *
 *  Copyright (C) 2002  The Exult Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef _DEBUGSERVER_H
#define _DEBUGSERVER_H

#if (defined(USE_EXULTSTUDIO) && defined(USECODE_DEBUGGER))

void Handle_client_debug_message(int &fd);
void Handle_debug_message(unsigned char *data, int datalen);

#endif
#endif
