// stdafx.h : include file for standard system include files,
//  or project specific include files that are used frequently, but
//      are changed infrequently
//

#if !defined(AFX_STDAFX_H__07A411AF_2AA7_43BB_AF36_B5999C1DEE6B__INCLUDED_)
#define AFX_STDAFX_H__07A411AF_2AA7_43BB_AF36_B5999C1DEE6B__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

// This MUST BE FIRST!
#include "msvc_kludges.h"

// Insert your headers here
#define WIN32_LEAN_AND_MEAN     // Exclude rarely-used stuff from Windows headers

#include <windows.h>

// TODO: reference additional headers your program requires here

//{{AFX_INSERT_LOCATION}}
// Microsoft Visual C++ will insert additional declarations immediately before the previous line.

#endif // !defined(AFX_STDAFX_H__07A411AF_2AA7_43BB_AF36_B5999C1DEE6B__INCLUDED_)
